package dmc;

/*
 * JChatPanel. java
 * Matthew Wronka <wronkm@rpi.edu>
 *
 * Chat UI.  Provides input method and
 * display area for output.
 * Copyright (C) 2002 Matthew Wronka <wronkm@rpi.edu>
 * This file is released under the GNU General Public License version 2,
 * a version of which should have been provided along with this file.
 *
 * You might also find a copy at:
 * http://dmc.sf.net/files/gpl.txt
 * http://www.fsf.org/copyleft/gpl.html
 */

import javax.swing.*;
import javax.swing.text.*;
import javax.swing.tree.*;

import java.awt.*;
import java.awt.event.*;
import java.awt.event.KeyEvent.*;
import java.util.Vector;

import wtalk.WTalk;

/**
 * Provides a GUI for communicating.
 * Both input, in the form of a JTextField,  
 *	and output, in the orm of a JTextPane,
 *	are provided.
 *
 * @author Matthew Wronka <wronkm@rpi.edu>
 * @version 0
 */
public class JChatPanel extends JPanel {

	static final public int SYSTEM_TAB	= 0;
	static final public int IM_TAB   	= 1;
	static final public int CHAT_TAB	= 2;

	final protected JTextPane textChat = new JTextPane();
	final protected JScrollPane chatScroll = new JScrollPane(textChat);
	final protected JTextField textInput = new JTextField();
	protected WTalk wt = null;
	protected ChatInputHandler chi = null;
	protected String CurrentRoom = null;
	private TabbedChat m_TabbedChat = null;

	private Vector txtHistory = new Vector();
	private Vector tmpHistory  = new Vector();
	private int indxHistory = -1;

	/**
	 * ctor which sets the current target, parent TabbedChat, and
	 * 	WTalk object.
	 *
	 * @param c	Parent TabbedChat.
	 * @param w	WTalk object to use.
	 * @param room  The target to which this JChatPanel communicates with.
	 */
	public JChatPanel(TabbedChat c, WTalk w, String room) {
		super();
		m_TabbedChat=c;
	
		chatScroll.setFocusable(false);	// whatever.
		textChat.setFocusable(true);	// Copy/Paste
		textInput.setFocusable(true);	// duh

		textChat.setEditable(false);
		chatScroll.setVerticalScrollBarPolicy( JScrollPane.VERTICAL_SCROLLBAR_ALWAYS);
		chatScroll.setMinimumSize(new Dimension(10, 10));

		initTextStyles(textChat);
		//InitReaders(textChat);

		chi = new ChatInputHandler(CurrentRoom=room, textChat, c, wt=w);

		textInput.addActionListener(new ActionListener(){
			public void actionPerformed(ActionEvent e){
				String s;
				Document d = textInput.getDocument();
				int len = d.getLength();

				try {
					s = d.getText(0, len);
					chi.doInput(s);
					d.remove(0, len);
					if(!s.equals("")) txtHistory.addElement(s);
					indxHistory=txtHistory.size();
					tmpHistory = (Vector)txtHistory.clone();
					tmpHistory.addElement("");
					//textChat.scrollRectToVisible(new Rectangle(0, textChat.getHeight(), 1, 1));
				} catch (BadLocationException ble){}
			}
		});

		//setBorder(BorderFactory.createEmptyBorder( 10, 10, 10, 10);
		setLayout(new BorderLayout());
		add(chatScroll, BorderLayout.CENTER);
		add(textInput, BorderLayout.SOUTH);
	
		addKeyBindings();
	}

	/**
	 * Replaces all text in Document d with String s.
	 */
	private void replaceText(Document d, String s){
		try{
			d.remove(0, d.getLength());
			d.insertString(0, s, null);
		}catch(Exception e){}

	}

        /**
         * Adds keybindigns for history
         */
        private void addKeyBindings() {
                textInput.getInputMap(WHEN_FOCUSED).
			put(KeyStroke.getKeyStroke("pressed DOWN"), 
				"next-command");
                textInput.getInputMap(WHEN_FOCUSED).
			put(KeyStroke.getKeyStroke("pressed UP"), 
				"prev-command");

                Action PrevCommand= new AbstractAction(){
                        public void actionPerformed(ActionEvent e){
				Document d = textInput.getDocument();
				try{tmpHistory.set(indxHistory, d.getText(0, d.getLength()));}catch(Exception ex){};
				if(--indxHistory >= 0)
					replaceText(d, (String)tmpHistory.get(indxHistory));
				else
					indxHistory=0;
                        }
                };

                Action NextCommand= new AbstractAction(){
                        public void actionPerformed(ActionEvent e){
				Document d = textInput.getDocument();
				try{tmpHistory.set(indxHistory, d.getText(0, d.getLength()));}catch(Exception ex){};
				if(++indxHistory < tmpHistory.size() )
					replaceText(d, (String)tmpHistory.get(indxHistory));
				else{
					indxHistory=tmpHistory.size()-1;
				}
			}
		};


		textInput.getActionMap().put("next-command", NextCommand);
		textInput.getActionMap().put("prev-command", PrevCommand);
	}


	/**
	 * Returns the JTextPane belonging to this JChatPane.
	 * <p>
	 * This may  be used to access the document properties, etc.
	 * for writing to the chat window.
	 *
	 * @return This JChatPane's JTextPane, which contains the active
	 *	conversation.
	 */
	public JTextPane getJTextPane(){
		return textChat;
	}

	/**
	 * Causes the JTextField which collects input to request focus.
	 */
	 public void requestFocus(){
		System.err.println(textInput.requestFocusInWindow());
	 }

	/**
	 * Initializes text styles for the textPane.
	 *
	 * @param textPane Textpane to which to apply the styles.	
	 */
	protected static void initTextStyles(JTextPane textPane) {
		// Initizlize some default styels.
		Style def = StyleContext.getDefaultStyleContext().
			getStyle(StyleContext.DEFAULT_STYLE);

		Style regular  = textPane.addStyle("regular", def);
		StyleConstants.setFontFamily(def, "SansSerif");

		Style s = null;

		s = textPane.addStyle("client", regular);
		StyleConstants.setBold(s, true);
		StyleConstants.setForeground(s, DmcColors.FONT_CLIENT);

		s = textPane.addStyle("wall", regular);
		StyleConstants.setBold(s, true);
		StyleConstants.setForeground(s, DmcColors.FONT_WALL);

		s = textPane.addStyle("server", regular);
		StyleConstants.setBold(s, true);
		//StyleConstants.setItalic(s, true);
		StyleConstants.setForeground(s, DmcColors.FONT_SERVER);

		s = textPane.addStyle("private", regular);
		StyleConstants.setItalic(s, true);
		StyleConstants.setForeground(s, DmcColors.FONT_PRIVATE);

		s = textPane.addStyle("chat", regular);
		//StyleConstants.setForeground(s, DmcColors.FONT_CHAT);

		s = textPane.addStyle("isay", regular);
		StyleConstants.setBold(s, true);
		StyleConstants.setForeground(s, DmcColors.FONT_ISAY);

		s = textPane.addStyle("usay", regular);
		StyleConstants.setBold(s, true);
		StyleConstants.setForeground(s, DmcColors.FONT_USAY);

		s = textPane.addStyle("auto", regular);
		StyleConstants.setBold(s, true);
		StyleConstants.setForeground(s, DmcColors.FONT_AUTO);

	}

	/**
	 * Returns the ChatInputHandler which negotiates encryption
	 * and executes internal commands.
	 *
	 * @return Returns this JChatPanel's ChatInputHandler.
	 */
	public ChatInputHandler getChatInputHandler() {
		return chi;
	}

	/**
	 * Returns the parent TabbedChat
	 *
	 * @return Conainer TabbedChat
	 */
	public TabbedChat getTabbedChat() {
		return m_TabbedChat;
	}

	/**
	 * Executes a command.
	 *
	 * @param command Command to execute.
	 */
	public void doCommand(String command) {
		chi.doCommand(command);
	}
}
