/***
 * General class for creating messages.  
 * Used to construct outgoing, parse incoming.
 */



package wtalk.protocols.irc;

import java.io.*;
import java.net.*;
import java.util.Vector;
import java.util.StringTokenizer;

public class IRCMessage {
	private String sPrefix = "";
	private String sCommand = "";
	private String sText = "";
	private String sCTCPCommand = "";
	private int iResponse = 0;	
	private Vector sParams = new Vector();
	private boolean bCTCP = false;

	public String toString() {
		String sMessage=sCommand;
		if(sPrefix!="") sMessage=":"+sPrefix+" " + sMessage;
		for(int i=0; i<sParams.size(); ++i)
			sMessage+= " " + (String)sParams.get(i);

		return sMessage;
	}

	public IRCMessage() {}
	public IRCMessage(String sMessage) { Parse(sMessage); }
	private String flattenParams() {
		String rval="";
		for(int i=0; i<sParams.size();++i){
			if(i>0) rval+= " ";
			rval+=(String)sParams.get(i);
		}

		return rval;
	}

	/*
	 * Should be used to handle received messages
	 */
	public void  Parse(String sMessage) {
		StringTokenizer st = new StringTokenizer(sMessage," ");
		if(sMessage.length() <=0) return;

		// Check for prefix
		if(sMessage.charAt(0)==':') { // get the prefix
			sPrefix = (st.nextToken()).substring(1);
		}

		// Get the command
		if(st.hasMoreTokens())
			sCommand = st.nextToken();

		try {
			iResponse = java.lang.Integer.parseInt(sCommand);
		} catch(NumberFormatException E) {
			iResponse = 0;
		}

		// Gather in all parameters
		// technically, there should be at most 15, but we don't care
		while(st.hasMoreTokens()) {
			String t=st.nextToken();
			if(t.length() <= 0) return;
			if(t.charAt(0)==':') { 
				t = t.substring(1);
				while(st.hasMoreTokens()) 
					t+=" " + st.nextToken();
			}

			sParams.addElement(t);
		}

		// Get the "text" into its own variable
		sText = parseText();



		// Deal with CTCP here
		if(isChatMsg()) {
			String s = sText;
			if(s.length() <= 0) return;
			if('\001' == s.charAt(0) && 
					'\001'  == s.charAt(s.length()-1)) {

				bCTCP = true;

				st = new StringTokenizer(sText.substring
						(1, s.length()-1), " ");

				if(st.hasMoreTokens()) {
					sCTCPCommand = st.nextToken();
				}

//				System.err.println("Command: " + sCTCPCommand +
//					"\tSTEXT: " + sText);

				sText = "";
				while(st.hasMoreTokens()) {
					sText += " " + st.nextToken();
				}

				sText = decodeCTCP(sText);
			}
		} 
	}

	static public String decodeCTCP(String s) {
		String rval="";
		int i;

		// "Low Level"
		for(i=0; i < s.length(); ++i ) {
			if(s.charAt(i)=='\020') {
				// Make sure someone isn't being a jerk
				if(i+1 < s.length())
					// Add next letter, inc i
					rval +=s.charAt(++i);
			} else 
				rval += s.charAt(i);

		}

		// "CTCP Level"
		s=rval; rval="";
		for(i=0; i < s.length(); ++i ) {
			if(s.charAt(i)=='\\') {
				if(i+1 < s.length()) {
					switch(s.charAt(i)) {
					case 'a':
						rval+='\001';
						break;
					case '\\':
						rval+='\\';
					}
				}
			} else {
				rval += s.charAt(i);
			}
		}

		return rval;
	}

	static public String encodeCTCP(String s) {
		String rval="";
		int i;

		// "CTCP Level"
		for(i=0; i< s.length(); ++i) {
			switch(s.charAt(i)) {
			case '\001':	// X-DELIM
				rval+="\\a";
				break;
			case '\\':	// X-QUOTE
				rval+= '\\';
			default:	// Fall through
				rval+= s.charAt(i);
			}
		}

		s=rval; rval="";

		// "Low Level"
		for(i=0; i< s.length(); ++i) {
			switch(s.charAt(i)) {
			case '\0':	// NUL
			case '\n':	// NL
			case '\r':	// CR
			case '\020':	// M-QUOTE
				rval+= '\020';
			default:	// Fall through
				rval+= s.charAt(i);
			}
		}


		return rval;

	}

	public String getSender() { 
		if(sCommand.equals("PING")) return sPrefix;

		StringTokenizer st = new StringTokenizer(sPrefix," !");

		if(st.hasMoreTokens())
			return st.nextToken();
		else return "";
	}

	public String getReceiver(){ 
		if(sParams.size()>0 && ( isChatMsg() || isStatusMsg() ) )
			return (String)sParams.get(0);

		// Server Replies
		if(sParams.size()==0) return "";

		switch(iResponse) {
		case IRCDefines.RPL_TOPIC:
		case IRCDefines.RPL_NOTOPIC:
		case IRCDefines.NS_RPL_ENDTOPIC:
		case IRCDefines.RPL_BANLIST:
		case IRCDefines.RPL_ENDOFBANLIST:
		case IRCDefines.RPL_TIME:
		case IRCDefines.RPL_AWAY:
			return (String)sParams.get(1);

		case IRCDefines.RPL_NAMEREPLY:
		case IRCDefines.RPL_ENDOFNAMES:
			if(((String)sParams.get(1)).equals("=")) 
				return (String)sParams.get(2);

			if(!((String)sParams.get(1)).equals("*"))
				return (String)sParams.get(1);
			else
				return "";
		default:
			return "";
		}
	}

	public String getCommand() { return sCommand; }

	public boolean isAutomatedMsg() { 
		return sCommand.equals("301");
	}

	public boolean isServerMsg() { 
		return !isAutomatedMsg() && !isStatusMsg() && !isChatMsg();

	}

	public boolean isChatMsg() {
		return  sCommand.equals("PRIVMSG") ||
			sCommand.equals("NOTICE") ;
	}

	public boolean isStatusMsg() {
		return	sCommand.equals("KICK") ||
			sCommand.equals("PART") ||
			sCommand.equals("JOIN") ||
			sCommand.equals("QUIT") ||
			sCommand.equals("MODE") ;
			
	}

	public boolean isCTCP() { return bCTCP; }

	public String getCTCPCommand() {
		return sCTCPCommand;
	}

	private String parseText() { 
		if(sCommand.equals("PING")) return flattenParams();
		if(sParams.size()>1){
			String rval="";	int i=1;

			// On a PRIVMSG, return the message
			if(sParams.size() >= 1 && 
				!(isServerMsg() ||isAutomatedMsg()))
				return (String)sParams.get(1);

		if(!getReceiver().equals("")) ++i;

			while(i<sParams.size()){
				rval += (String)sParams.get(i++) + " ";
			}

			return rval;
		} else
			return "";
	}

	public String getText() {
		return sText;
	}

	public String getParam(int i) {
		return (String)sParams.get(i);
	}

	public int getType() { return 0;}

	/*
	 * Used with outgoing messages
	 */
	public void setCommand(String sCom) { sCommand = sCom; }
	public void addParam(String sParam) { sParams.addElement(sParam); }

	public void send(PrintWriter out) {
		out.println(toString());
	}
}
